package ch.epfl.ppo;

import static java.lang.Math.PI;
import static java.lang.Math.sin;

import java.util.function.UnaryOperator;

/**
 * Une animation, représentée par une fonction associant une image à chaque
 * instant.
 */
public interface Animation<T> {
    public Image<T> imageAt(double time);

    public static <T> Animation<T> constant(Image<T> image) {
        return t -> image;
    }

    public static Animation<ColorRGB> rainbow(double fR, double phiR, double fG, double phiG, double fB, double phiB) {
        double pi2 = 2 * PI;
        UnaryOperator<Double> r = t -> sin(t * fR * pi2 + phiR) / 2d + 0.5;
        UnaryOperator<Double> g = t -> sin(t * fG * pi2 + phiG) / 2d + 0.5;
        UnaryOperator<Double> b = t -> sin(t * fB * pi2 + phiB) / 2d + 0.5;
        return t -> Image.constant(new ColorRGB(r.apply(t), g.apply(t), b.apply(t)));
    }

    public static Animation<ColorRGB> compose(Animation<ColorRGB> bg, Animation<ColorRGB> fg, Animation<Double> m) {
        return t -> Image.compose(bg.imageAt(t), fg.imageAt(t), m.imageAt(t));
    }

    public static Animation<Double> dissolveMask(double t1, double t2) {
        if (! (0 <= t1 && t1 <= t2))
            throw new IllegalArgumentException();
        double dt = t2 - t1;
        return t -> Image.constant(Math.min(1, Math.max(t - t1, 0) / dt));
    }

    public default Animation<T> loop(double t1, double t2) {
        if (! (0 <= t1 && t1 < t2))
            throw new IllegalArgumentException("invalid time interval: [" + t1 + ";" + t2 + "]");
        return t -> imageAt(t1 + t % (t2 - t1));
    }
}