package ch.epfl.ppo;

import java.awt.image.BufferedImage;
import java.util.stream.Stream;

/**
 * Une image continue et infinie, représentée par une fonction associant une
 * couleur à chaque point du plan.
 */
public interface Image<T> {
    public T valueAt(double x, double y);

    public static final Image<ColorRGB> RED_CIRCLE = (x, y) -> {
        double r = Math.sqrt(x * x + y * y);
        return r <= 1.0 ? ColorRGB.RED : ColorRGB.WHITE;
    };

    public static <T> Image<T> constant(T value) {
        return (x, y) -> value;
    }

    public static Image<ColorRGB> chessboard(ColorRGB c1, ColorRGB c2, double w) {
        if (! (w > 0))
            throw new IllegalArgumentException("non-positive width: " + w);

        return (x, y) -> {
            double sqX = Math.floor(x / w), sqY = Math.floor(y / w);
            return ((int)(sqX + sqY)) % 2 == 0 ? c1 : c2;
        };
    }

    public static Image<ColorRGB> target(ColorRGB c1, ColorRGB c2, double w) {
        if (! (w > 0))
            throw new IllegalArgumentException("non-positive width: " + w);

        return (x, y) -> {
            double r = Math.sqrt(x * x + y * y);
            return (int)Math.floor(r / w) % 2 == 0 ? c1 : c2;
        };
    }

    public static Image<ColorRGB> linearHorizontalGradient(ColorRGB cL, double xL, ColorRGB cR, double xR) {
        if (! (xR > xL))
            throw new IllegalArgumentException("invalid interval "+ xL + " - " + xR);

        return (x, y) -> {
            if (x <= xL)
                return cL;
            else if (x >= xR)
                return cR;
            else {
                double p = (x - xL) / (xR - xL);
                return cL.mixWith(cR, p);
            }
        };
    }

    public static Image<Double> linearHorizontalGradientMask(double xL, double xR) {
        if (! (xR > xL))
            throw new IllegalArgumentException("invalid interval "+ xL + " - " + xR);

        return (x, y) -> {
            if (x <= xL)
                return 0d;
            else if (x >= xR)
                return 1d;
            else
                return (x - xL) / (xR - xL);
        };
    }

    public static Image<ColorRGB> ofDiscreteImage(ColorRGB bg, BufferedImage i) {
        double w = i.getWidth();
        double h = i.getHeight();
        double s = w / 2d;

        return (x, y) -> {
            int iX = (int)Math.round( x * s + w / 2d);
            int iY = (int)Math.round(-y * s + h / 2d);

            if (0 <= iX && iX < w && 0 <= iY && iY < h)
                return new ColorRGB(i.getRGB(iX, iY));
            else
                return bg;
        };
    }

    public static Image<ColorRGB> ofBooleanImage(Image<Boolean> m, ColorRGB f, ColorRGB t) {
        return (x, y) -> m.valueAt(x, y) ? t : f;
    }

    public static Image<Boolean> mandelbrotSet(int maxIt) {
        return (x, y) -> {
            Complex c = new Complex(x, y);
            return Stream.iterate(c, z -> z.squared().add(c))
                    .limit(maxIt - 1)
                    .allMatch(z -> z.modulusSquared() < 4d);
        };
    }

    public static Image<ColorRGB> compose(Image<ColorRGB> bg, Image<ColorRGB> fg, Image<Double> m) {
        return (x, y) ->
           bg.valueAt(x, y).mixWith(fg.valueAt(x, y), m.valueAt(x, y));
    }

    public default Image<T> flattened() {
        return (x, y) -> valueAt(x, y * 2d);
    }

    public default Image<T> rotated(double a) {
        double cosInvA = Math.cos(-a);
        double sinInvA = Math.sin(-a);
        return (x, y) -> {
            double rX = x * cosInvA - y * sinInvA;
            double rY = x * sinInvA + y * cosInvA;
            return valueAt(rX, rY);
        };
    }
}
